/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/graphics.h: Header for graphics functionality.
 */

/*
 * Coordinate transformation is performed as:
 *
 *     [X Y Z W] = (([x y z w] * M_model) * M_view) * M_projection
 *
 * (this is the same as the OpenGL fixed-function pipeline transformation,
 * rewritten for row-major matrices).  [x y z w] is a row vector containing
 * the input (model space) coordinates, and [X Y Z W] is a row vector
 * containing the output (device space) coordinates.  Device space is in
 * [-1,+1] for all three dimensions, with the positive axes pointing right,
 * up, and away from the viewer (a left-handed coordinate system).
 *
 * All color parameters to graphics functions are floating-point values
 * between 0.0 (minimum value) and 1.0 (maximum value) inclusive; on an
 * 8-bit color display, for example, 0.0 maps to color value 0 and 1.0
 * maps to color value 255.  Colors are specified by their red, green,
 * and blue components (usually named "r", "g", and "b"), sometimes with
 * an alpha ("a") component as well.  When specified using a vector value,
 * the red, green, blue, and alpha components map to x, y, z, and w
 * respectively.
 *
 * Unless otherwise specified, all graphics functions (including texture
 * and font functions) must be called from the program's main thread
 * (the thread on which sil_main() is called).
 *
 * Some functions which take vector or matrix input parameters can also be
 * called from C++ using a (non-pointer) value; see the "C++ wrappers"
 * section near the end of this file.
 */

#ifndef SIL_GRAPHICS_H
#define SIL_GRAPHICS_H

EXTERN_C_BEGIN

struct Matrix4f;
struct Vector2f;
struct Vector3f;
struct Vector4f;

/*************************************************************************/
/*********************** Data types and constants ************************/
/*************************************************************************/

/**
 * GraphicsDisplaySizeList:  Structure containing an array of display
 * sizes.  Returned from graphics_list_display_sizes().
 */
typedef struct GraphicsDisplaySizeList GraphicsDisplaySizeList;
struct GraphicsDisplaySizeList {
    /* Number of entries in the sizes[] array. */
    int num_sizes;
    /* Array of display sizes (in pixels). */
    struct {int width, height;} sizes[1];
};

/*-----------------------------------------------------------------------*/

/**
 * GraphicsBlendFactor:  Constants identifying a pixel multiplication
 * factor for blend operations.  The literal values "0" and "1" can be used
 * in place of GRAPHICS_BLEND_ZERO and GRAPHICS_BLEND_ONE respectively.
 *
 * In blending operations, the "source" pixel value refers to the new pixel
 * value being drawn to the framebuffer; for example, if a half-transparent
 * orange polygon is being drawn, the source pixel value would be
 * (R=1, G=0.5, B=0, A=0.5).  The "destination" pixel value refers to the
 * current value of the pixel in the framebuffer which will be affected by
 * the draw operation.
 *
 * Not all blend factors or combinations thereof are supported on all
 * systems.  In particular, the following may not be supported:
 *    - Color squaring (for example, using SRC_COLOR as the source factor)
 *    - Constant blend factors (CONSTANT, INV_CONSTANT)
 *
 * Note that, while destination alpha factors are included for completeness,
 * it is currently undefined whether framebuffers (including both the
 * display buffer and user-created framebuffers) include an alpha channel,
 * and thus the effect of using DEST_ALPHA and INV_DEST_ALPHA is undefined.
 */
enum GraphicsBlendFactor {
    GRAPHICS_BLEND_ZERO = 0,       // out = 0
    GRAPHICS_BLEND_ONE,            // out = in
    GRAPHICS_BLEND_SRC_COLOR,      // out = in * src
    GRAPHICS_BLEND_SRC_ALPHA,      // out = in * src.a
    GRAPHICS_BLEND_INV_SRC_ALPHA,  // out = in * (1-src.a)
    GRAPHICS_BLEND_DEST_COLOR,     // out = in * dest
    GRAPHICS_BLEND_DEST_ALPHA,     // out = in * dest.a
    GRAPHICS_BLEND_INV_DEST_ALPHA, // out = in * (1-dest.a)
    GRAPHICS_BLEND_CONSTANT,       // out = in * blend_color
    GRAPHICS_BLEND_INV_CONSTANT,   // out = in * (1-blend_color)
};
typedef enum GraphicsBlendFactor GraphicsBlendFactor;

/**
 * GraphicsBlendOperation:  Constants identifying a pixel combination
 * function for blend operations.
 *
 * Note that some (very old) systems may not support functions other than
 * GRAPHICS_BLEND_ADD.
 */
enum GraphicsBlendOperation {
    GRAPHICS_BLEND_ADD = 1,     // out = src*srcFactor + dest*destFactor
    GRAPHICS_BLEND_SUB,         // out = src*srcFactor - dest*destFactor
    GRAPHICS_BLEND_RSUB,        // out = dest*destFactor - src*srcFactor
};
typedef enum GraphicsBlendOperation GraphicsBlendOperation;

/**
 * GraphicsComparisonType:  Constants identifying comparison types for
 * rendering test operations.  See the documentation of each function
 * using these constants for details.
 */
enum GraphicsComparisonType {
    GRAPHICS_COMPARISON_FALSE = 0,
    GRAPHICS_COMPARISON_TRUE,
    GRAPHICS_COMPARISON_EQUAL,
    GRAPHICS_COMPARISON_NOT_EQUAL,
    GRAPHICS_COMPARISON_LESS,
    GRAPHICS_COMPARISON_LESS_EQUAL,
    GRAPHICS_COMPARISON_GREATER_EQUAL,
    GRAPHICS_COMPARISON_GREATER,
};
typedef enum GraphicsComparisonType GraphicsComparisonType;

/**
 * GraphicsFaceCullMode:  Constants identifying modes for face (polygon)
 * culling, used with graphics_set_face_cull().
 */
enum GraphicsFaceCullMode {
    GRAPHICS_FACE_CULL_NONE = 0,  // Do not cull any polygons.
    GRAPHICS_FACE_CULL_CW,        // Cull when vertices are in clockwise order.
    GRAPHICS_FACE_CULL_CCW,       // Cull when vertices are counterclockwise.
};
typedef enum GraphicsFaceCullMode GraphicsFaceCullMode;

/**
 * GraphicsPrimitiveID:  Constants identifying primitive types.  The types
 * defined here are treated identically to the corresponding OpenGL
 * primitives of the same names.
 */
enum GraphicsPrimitiveID {
    GRAPHICS_PRIMITIVE_POINTS = 1,
    GRAPHICS_PRIMITIVE_LINES,
    GRAPHICS_PRIMITIVE_LINE_STRIP,
    GRAPHICS_PRIMITIVE_TRIANGLES,
    GRAPHICS_PRIMITIVE_TRIANGLE_STRIP,
    GRAPHICS_PRIMITIVE_QUADS,
    GRAPHICS_PRIMITIVE_QUAD_STRIP,
};
typedef enum GraphicsPrimitiveID GraphicsPrimitiveID;

/**
 * GraphicsStencilOp:  Constants identifying stencil operations, used with
 * graphics_set_stencil_operations().
 */
enum GraphicsStencilOp {
    GRAPHICS_STENCIL_KEEP = 1,  // Leave the current value unchanged.
    GRAPHICS_STENCIL_CLEAR,     // Clear the value to zero.
    GRAPHICS_STENCIL_REPLACE,   // Set the value to the comparison reference.
    GRAPHICS_STENCIL_INCR,      // Increment the value (clamped to maximum).
    GRAPHICS_STENCIL_DECR,      // Decrement the value (clamped to zero).
    GRAPHICS_STENCIL_INVERT,    // Bitwise invert the value.
};
typedef enum GraphicsStencilOp GraphicsStencilOp;

/**
 * GraphicsVertexDataID:  Constants identifying various kinds of data that
 * can be stored in vertex data streams.  The constant names take the
 * following form:
 *     <data>_<count><format>
 * where <data> is the data content (POSITION, TEXCOORD, etc.), <count> is
 * the number of data elements (e.g. 3 for a 3D vertex position), and
 * <format> identifies the numeric format of the data:
 *     F = 32-bit float, H = 16-bit float, D = 64-bit float,
 *     I = 32-bit integer, S = 16-bit integer, B = 8-bit integer,
 *     UB = 8-bit unsigned integer
 * much like the suffixes used in OpenGL function names.
 *
 * For ease of debugging, the constant values are defined using the
 * following scheme:
 *     0xABCD
 *       |||`------- Number of components (e.g. "3" for a 3-component value)
 *       ||`------ Data type: (is_float ? 8 : 0) | (log2(num_bits/8))
 *       ||           e.g. 32-bit float = 8 | log2(32/8) = 0xA
 *       |`----- Minor grouping (e.g. different kinds of color values)
 *       `---- Major grouping (e.g. position, texture coordinate, color...)
 * However, both callers and implementations MUST NOT rely on this format
 * as it may change in the future.
 *
 * Note that callers of graphics_create_primitive() should _not_ use these
 * constants directly, but should instead use the GRAPHICS_VERTEX_FORMAT()
 * macro (see below) to generate the values to store in the "format" array.
 */
enum GraphicsVertexDataID {
    GRAPHICS_VERTEX_POSITION_2S = 0x0012,
    GRAPHICS_VERTEX_POSITION_2F = 0x00A2,
    GRAPHICS_VERTEX_POSITION_3F = 0x00A3,
    GRAPHICS_VERTEX_POSITION_4F = 0x00A4,

    GRAPHICS_VERTEX_TEXCOORD_2F = 0x10A2,

    GRAPHICS_VERTEX_COLOR_4UB = 0x2004,
    GRAPHICS_VERTEX_COLOR_4F  = 0x20A4,
};
typedef enum GraphicsVertexDataID GraphicsVertexDataID;

/**
 * GRAPHICS_VERTEX_FORMAT:  Macro for constructing values to store into a
 * "format" array for passing to graphics_create_primitive() or
 * sys_graphics_create_primitive().  The "format" array should consist of
 * a sequence of values generated by this macro, followed by the value
 * zero to terminate the array.
 *
 * [Parameters]
 *     type: Data type (a GRAPHICS_VERTEX_* constant, but with
 *         "GRAPHICS_VERTEX_" omitted from the constant name).
 *     offset: Offset of data element within vertex data structure, in
 *         bytes (must be nonnegative and <=65535).
 * [Return value]
 *     Value to store in "format" array.
 */
#define GRAPHICS_VERTEX_FORMAT(type,offset) \
    ((uint32_t)(GRAPHICS_VERTEX_##type << 16 | ((offset) & 0xFFFF)))

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/*------------- Initialization, configuration, and cleanup --------------*/

/**
 * graphics_device_width, graphics_device_height:  Return the current width
 * or height of the display device.  This will typically be constant for a
 * particular runtime environment, but on systems which allow the device
 * resolution to be changed at runtime (such as PC operating systems), the
 * return values from these functions will change to reflect the current
 * resolution at the time the functions are called.
 *
 * Note that when using windowed mode, passing these values to
 * graphics_set_display_mode() will not necessarily result in a window
 * which covers the entire display device!  For example, if the system
 * has a working area (such as a desktop) which is larger than the
 * current display resolution, the system may place the window partly or
 * entirely offscreen.  On the flip side, setting a fullscreen display
 * mode using these values will always succeed (assuming no runtime
 * failures such as insufficient memory).
 *
 * These functions may be called from any thread.
 *
 * [Return value]
 *     Display device width or height, in pixels.
 */
extern int graphics_device_width(void);
extern int graphics_device_height(void);

/**
 * graphics_has_windowed_mode:  Return whether the system supports the
 * concept of a windowed mode in which a display region of arbitrary size
 * (up to the screen size) can be used.
 *
 * This function may be called from any thread.
 *
 * [Return value]
 *     True if windowed mode is supported, false if not.
 */
extern int graphics_has_windowed_mode(void);

/**
 * graphics_list_display_sizes:  Return an array containing display sizes
 * (i.e., resolutions) supported by the system.  For systems which support
 * the concept of a windowed mode, this is the list of display sizes
 * supported in fullscreen mode.  Returned values are ordered from the
 * smallest to the largest number of total pixels.
 *
 * [Return value]
 *     List of supported display sizes.  The return value will never be NULL.
 */
extern const GraphicsDisplaySizeList *graphics_list_display_sizes(void);

/**
 * graphics_set_display_attr:  Set a system-specific display attribute.
 *
 * The following attributes are defined (though not necessarily supported)
 * for all systems:
 *
 * "depth_bits": 1 value of type int
 *     Sets the minimum number of bits per pixel to use for the depth
 *     buffer.  The actual depth buffer resolution may be greater than
 *     this value, depending on available hardware display modes.  Even
 *     if this value is successfully set, graphics_set_display_mode() will
 *     fail if no display mode is available that supports at least the
 *     requested number of bits.  A value of zero indicates that a depth
 *     buffer is not required, and the system may choose not to allocate a
 *     depth buffer (in which case any operations involving the depth
 *     buffer are undefined).  The default value is 16.
 *
 *     This value is also applied to all user-created framebuffers.
 *
 * "multisample": 1 value of type int
 *     Sets the number of samples to use for multisample antialiasing
 *     (MSAA).  A value of 1 disables MSAA.  The default value is 1.
 *
 *     This value is NOT applied to user-created framebuffers; there is
 *     currently no way to enable multisampling on such framebuffers.
 *
 * "preserve_state": 1 value of type int
 *     Sets whether to preserve graphics state (including texture and
 *     primitive data) across a window size change.  A nonzero value
 *     enables preserving state; a zero value disables it.  If disabled,
 *     the client code is responsible for resetting render state and
 *     destroying and recreating all textures and primitives after calling
 *     graphics_set_display_mode() to change the display parameters.  The
 *     default is to preserve graphics state if possible.
 *
 *     Some systems may preserve graphics state in some or all cases, or
 *     may not support multiple display modes to begin with; on such
 *     systems, the value of this attribute is ignored.  Conversely, some
 *     systems may be unable to preserve graphics state; on such systems,
 *     this attribute's value will always be false, and attempting to set
 *     it to true will fail.
 *
 *     Note that preserving state may have a negative performance impact
 *     if the state preservation logic is actually invoked.  In particular,
 *     OpenGL texture data is currently recreated as 32-bit RGBA regardless
 *     of its original format; this may lead to excessive memory usage and
 *     reduced performance after a display mode change.
 *
 * "stencil_bits": 1 value of type int
 *     Sets the minimum number of bits per pixel to use for the stencil
 *     buffer.  The actual stencil buffer resolution may be greater than
 *     this value, depending on available hardware display modes.  Even
 *     if this value is successfully set, graphics_set_display_mode() will
 *     fail if no display mode is available that supports at least the
 *     requested number of bits.  A value of zero indicates that a stencil
 *     buffer is not required, and the system may choose not to allocate a
 *     stencil buffer (in which case any operations involving the stencil
 *     buffer are undefined).  The default value is 0.
 *
 *     This value is also applied to all user-created framebuffers.
 *
 * "vsync": 1 value of type int
 *     Sets whether to synchronize the beginning of each frame with a
 *     periodic external signal such as the display refresh.  A nonzero
 *     value enables synchronization, meaning that graphics_start_frame()
 *     may block; zero disables synchronization, allowing frames to be
 *     drawn as fast as possible but increasing the possibility of tearing
 *     or other visual artifacts.  Attempting to set a nonzero value on
 *     systems which do not support synchronization, or attempting to set
 *     zero on systems which require synchronization, will fail.  The
 *     default is to enable synchronization if supported.
 *
 * "window": 1 value of type int
 *     Sets whether graphics_set_display_mode() will use a windowed or
 *     full-screen mode.  Passing a nonzero value selects windowed mode;
 *     zero selects full-screen mode.  On systems which do not support the
 *     concept of a windowed mode, attempting to set a nonzero value will
 *     fail (setting zero will always succeed on such systems).  The
 *     default is to use windowed mode if available, else full-screen mode.
 *
 * Other attribute names and values are system-dependent; it is the
 * responsibility of the caller to pass correct values for the chosen
 * attribute.
 *
 * After calling this function, the state of the display is undefined
 * until the next graphics_set_display_mode() call.
 *
 * [Parameters]
 *     name: Attribute name.
 *     ...: Attribute value(s).
 * [Return value]
 *     True if the display attribute was successfully set; false if the
 *     attribute is unknown or an error occurred.
 */
extern int graphics_set_display_attr(const char *name, ...);

/**
 * graphics_set_display_mode:  Initialize the display to the given size,
 * using display attributes set through graphics_set_display_attr().  This
 * function must not be called between graphics_start_frame() and
 * graphics_finish_frame().
 *
 * If the display attribute "preserve_state" is enabled (as it is by
 * default), all graphics state, including previously-created textures
 * and primitives, is preserved across successful display mode changes.
 * If "preserve_state" is disabled, the state after a successful display
 * mode change is implementation-dependent.
 *
 * If this function fails, the state of the display is left undefined.
 *
 * [Parameters]
 *     width, height: Display size, in pixels.
 * [Return value]
 *     True if the display mode was successfully set; false on error.
 */
extern int graphics_set_display_mode(int width, int height);

/**
 * graphics_renderer_info:  Return a string describing the low-level
 * rendering backend.  In environments which use OpenGL, this will
 * typically be an OpenGL version string.
 *
 * This function may only be called after a display mode has been set.
 * Calling this function at other times results in undefined behavior.
 *
 * [Return value]
 *     Renderer information string.
 */
extern const char *graphics_renderer_info(void);

/**
 * graphics_display_is_window:  Return whether the current display mode
 * is a windowed mode.  On systems which do not support the concept of a
 * windowed mode, this function will always return false.
 *
 * [Return value]
 *     True if the display mode is a windowed mode; false if the display
 *     mode is a full-screen mode or if no display mode is set.
 */
extern int graphics_display_is_window(void);

/**
 * graphics_set_window_title:  Set the title to be displayed in the window
 * title bar (when in windowed mode) or any equivalent system facility,
 * such as a taskbar or application list.
 *
 * [Parameters]
 *     title: Title string.
 */
extern void graphics_set_window_title(const char *title);

/**
 * graphics_set_window_icon:  Set the icon to be associated with the
 * window, if relevant for the system.  The texture must not be locked,
 * and its pixel data must be readable.  The pixel data is copied out of
 * the texture, so the texture may be safely modified or destroyed after
 * this function returns.
 *
 * On Microsoft Windows, the texture must be 32x32 pixels in size.
 *
 * Note that it is not currently possible to set an icon before opening a
 * window, since OpenGL (which is used for texture management) will not be
 * initialized.  Attempting to call this function before calling
 * graphics_set_display_mode() may cause the program to crash.
 *
 * [Parameters]
 *     texture: Texture to use as icon image.
 */
extern void graphics_set_window_icon(int texture);

/**
 * graphics_show_mouse_pointer:  Set whether to display a system-level
 * visible mouse pointer on top of the display, for systems which have
 * such a concept.  Some systems may not allow the pointer to be hidden.
 *
 * By default, the mouse pointer (if any) is hidden if possible.
 *
 * [Parameters]
 *     on: True to display the pointer, false to hide it.
 */
extern void graphics_show_mouse_pointer(int on);

/**
 * graphics_get_mouse_pointer_state:  Return whether the system-level
 * mouse pointer display is enabled.
 *
 * [Return value]
 *     True if the pointer is displayed, false if not.
 */
extern int graphics_get_mouse_pointer_state(void);

/**
 * graphics_display_width, graphics_display_height:  Return the current
 * width or height of the display.  These values will never change while
 * a frame is being rendered (between graphics_start_frame() and
 * graphics_finish_frame()).
 *
 * These functions may be called from any thread.
 *
 * [Return value]
 *     Display width or height, in pixels, or zero if the display mode
 *     has not been set.
 */
extern PURE_FUNCTION int graphics_display_width(void);
extern PURE_FUNCTION int graphics_display_height(void);

/**
 * graphics_frame_period:  Return the nominal frame period of the current
 * display mode, in seconds per frame, or zero if the frame period is
 * unknown.
 *
 * [Return value]
 *     Frame period, in seconds per frame, or zero if unknown.
 */
extern double graphics_frame_period(void);

/**
 * graphics_has_focus:  For systems with a concept of input focus (typical
 * PC windowing systems, for example), return whether this program
 * currently has the input focus.  On other systems, this function always
 * returns true.
 *
 * [Return value]
 *     True if this program has the input focus, false if not.
 */
extern int graphics_has_focus(void);

/*--------------- Frame start/finish and synchronization ----------------*/

/**
 * graphics_start_frame:  Start rendering a frame.  This function must be
 * called before any functions which operate on the rendering target,
 * specifically:
 *    - graphics_bind_framebuffer()
 *    - graphics_clear()
 *    - graphics_draw_indexed_vertices()
 *    - graphics_draw_primitive()
 *    - graphics_draw_vertices()
 *    - graphics_end_and_draw_primitive()
 *    - graphics_read_pixels()
 *
 * If the "vsync" display attribute is enabled, this function may block
 * for up to graphics_frame_period() seconds.
 *
 * After this call, the display framebuffer is bound to the rendering
 * target, and the contents of the display framebuffer are undefined.
 */
extern void graphics_start_frame(void);

/**
 * graphics_finish_frame:  Finish rendering a frame.  No rendering functions
 * (see graphics_start_frame()) may be called after this function until the
 * next call to graphics_start_frame().
 */
extern void graphics_finish_frame(void);

/**
 * graphics_sync:  Wait for any background rendering operations to complete.
 */
extern void graphics_sync(void);

/**
 * graphics_flush_resources:  Immediately release any resources which are
 * pending deletion.  This function implicitly performs a graphics_sync()
 * operation, so it should not be called from performance-critical code.
 */
extern void graphics_flush_resources(void);

/*-------------------- Display clearing and reading ---------------------*/

/**
 * graphics_clear:  Clear the display or currently bound framebuffer to
 * the specified color, and if depth buffer writing is enabled, clear the
 * depth buffer to empty (all pixels set to the farthest depth).
 *
 * This function ignores the viewport region setting, but honors the clip
 * region setting.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     r, g, b: Clear color.
 */
extern void graphics_clear(float r, float g, float b);

/**
 * graphics_read_pixels:  Read pixel data from the display or currently
 * bound framebuffer.  If reading from a double-buffered display, this
 * function returns pixels from the current (in-progress) frame, not the
 * previous (displayed) frame.
 *
 * It is not an error to specify a region partly or completely outside
 * the display or framebuffer, but the out-of-bounds portion of the
 * returned buffer will be undefined.
 *
 * The returned pixel data will be in bottom-to-top, left-to-right order,
 * using RGBA format with 1 byte per pixel (the alpha byte will always be
 * 255).
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     x, y: Pixel coordinates of upper-left corner of region to read
 *         (0,0 is the bottom-left corner of the display).
 *     w, h: Size of region to read, in pixels.
 *     buffer: Buffer into which pixel data will be stored.  The buffer
 *         must have room for at least w*h*4 bytes.
 * [Return value]
 *     True on success, false on error.
 */
extern int graphics_read_pixels(int x, int y, int w, int h, void *buffer);

/*----------------------- Framebuffer management ------------------------*/

/**
 * graphics_framebuffer_supported:  Return whether the platform supports
 * offscreen framebuffers.
 *
 * [Return value]
 *     True if offscreen framebuffers are supported, false if not.
 */
extern int graphics_framebuffer_supported(void);

/**
 * graphics_create_framebuffer:  Create a new offscreen framebuffer.
 *
 * [Parameters]
 *     width, height: Framebuffer size, in pixels.
 * [Return value]
 *     Framebuffer ID (nonzero), or zero on error.
 */
extern int graphics_create_framebuffer(int width, int height);

/**
 * graphics_destroy_framebuffer:  Destroy an offscreen framebuffer.  Does
 * nothing if framebuffer_id == 0.
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer to destroy.
 */
extern void graphics_destroy_framebuffer(int framebuffer_id);

/**
 * graphics_framebuffer_width, graphics_framebuffer_height:  Return the
 * given framebuffer's width or height.
 *
 * These functions may be called from any thread.
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer for which to retrieve size.
 * [Return value]
 *     Width or height of framebuffer, in pixels.
 */
extern PURE_FUNCTION int graphics_framebuffer_width(int framebuffer_id);
extern PURE_FUNCTION int graphics_framebuffer_height(int framebuffer_id);

/**
 * graphics_bind_framebuffer:  Bind the given framebuffer to the current
 * rendering target.  If framebuffer_id == 0, cancel any framebuffer
 * binding so that the display is bound to the rendering target.
 *
 * Note that it is usually necessary to reset the render viewport with
 * graphics_set_viewport() after changing to a framebuffer of a different
 * size.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer to bind, or zero to cancel binding.
 */
extern void graphics_bind_framebuffer(int framebuffer_id);

/**
 * graphics_apply_framebuffer:  Apply the given framebuffer as a texture
 * for rendering.  If the framebuffer is currently bound to the rendering
 * target, or is subsequently bound to the rendering target before another
 * texture is applied, the results are undefined.
 *
 * To cancel the application of a framebuffer, call texture_apply(0).
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer to apply.
 */
extern void graphics_apply_framebuffer(int framebuffer_id);

/**
 * graphics_set_framebuffer_antialias:  Set whether the framebuffer should
 * be antialiased when used as a texture.  Defaults to true (antialiasing
 * enabled).
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer to modify.
 *     on: True to enable antialiasing, false to disable.
 */
extern void graphics_set_framebuffer_antialias(int framebuffer_id, int on);

/**
 * graphics_discard_framebuffer_data:  Discard all pixel data in the given
 * framebuffer.  This function serves as a hint to the graphics hardware
 * that the given framebuffer's contents do not need to be stored to
 * memory.  On some systems, calling this function when a framebuffer is
 * no longer needed (for example, after drawing the framebuffer to the
 * display) can improve performance.
 *
 * Calling this function has no effect on renderer output as long as no
 * attempt is made to read from the framebuffer before the next time it is
 * written to.  In particular, the framebuffer does not need to be unbound
 * before calling this function.
 *
 * [Parameters]
 *     framebuffer_id: ID of framebuffer whose contents may be discarded.
 */
extern void graphics_discard_framebuffer_data(int framebuffer_id);

/*------------------ Primitive creation and rendering -------------------*/

/**
 * graphics_begin_primitive:  Create a new graphics primitive object with
 * no attached vertex data.  Vertices can be subsequently added with
 * graphics_add_vertex(); the parameters supplied for the first vertex
 * will determine the vertex data format.  The primitive cannot be drawn
 * until it has been finalized with graphics_end_primitive().
 *
 * Callers may safely ignore the return value of this function and use
 * the return value of graphics_{end,end_and_draw}_primitive() to check
 * the success or failure of the primitive creation operation.  If this
 * function fails, all subsequent graphics_add_vertex() calls will be
 * ignored, and graphics_{end,end_and_draw}_primitive() will return zero
 * to indicate failure.
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*).
 * [Return value]
 *     True on success, false on error.
 */
extern int graphics_begin_primitive(GraphicsPrimitiveID type);

/**
 * graphics_add_vertex:  Add a vertex to the vertex data stream for a
 * primitive object created with graphics_begin_primitive().  If this is
 * the first vertex for the primitive, the presence or absence of the
 * vector parameters to this function will determine the data format;
 * otherwise, the presence or absence of the parameters must match that
 * of the first vertex, or an error will be signalled and the primitive
 * aborted.
 *
 * The format of vertex data which can be constructed through this function
 * is limited, and may not be optimized for fast rendering.  If the vertex
 * data is known in advance, it is generally preferable to create the
 * primitive object ahead of time with graphics_create_primitive().
 *
 * Position data for the vertex is required (the function will fail if
 * position == NULL).  If a texture has been applied with texture_apply()
 * but texture coordinates are not included in the vertex data, the
 * primitive will still be rendered but texturing behavior is undefined.
 *
 * As with graphics_begin_primitive(), callers may safely ignore the
 * return value of this function.  If this function fails for any reason
 * (including missing or extraneous data on the second or a later vertex),
 * all subsequent graphics_add_vertex() calls will be ignored, and
 * graphics_{end,end_and_draw}_primitive() will return zero to indicate
 * failure.
 *
 * [Parameters]
 *     position: Pointer to vector containing vertex position (required).
 *     texcoord: Pointer to vector containing texture coordinates (may be
 *         NULL).
 *     color: Pointer to vector containing color value (may be NULL).
 * [Return value]
 *     True on success, false on error.
 */
extern int graphics_add_vertex(const struct Vector3f *position,
                               const struct Vector2f *texcoord,
                               const struct Vector4f *color);

/**
 * graphics_end_primitive:  Mark the end of vertices for a primitive
 * object created with graphics_begin_primitive(), and return an ID for
 * the primitive object.
 *
 * [Return value]
 *     Primitive ID (nonzero), or zero on error.
 */
extern int graphics_end_primitive(void);

/**
 * graphics_end_and_draw_primitive:  Mark the end of vertices for a
 * primitive object created with graphics_begin_primitive(), then
 * immediately draw and destroy the primitive object.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Return value]
 *     True on success, false on error.
 */
extern int graphics_end_and_draw_primitive(void);

/**
 * graphics_create_primitive:  Create a new graphics primitive object from
 * a preexisting vertex data buffer.
 *
 * The "format" parameter describes the format of the vertex data, and points
 * to an array of uint32_t values generated by the GRAPHICS_VERTEX_FORMAT()
 * macro, terminated by a value of zero.  Unrecognized or unsupported format
 * values will cause an error.  Behavior is undefined if the array does not
 * include a POSITION_* format parameter; if a texture has been applied
 * with texture_apply() but the array does not include a TEXCOORD_* format
 * parameter, the primitive will be rendered but texturing behavior is
 * undefined (as with graphics_add_vertex()).
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*).
 *     data: Vertex data pointer.
 *     format: Vertex data format description.
 *     size: Size of a single vertex, in bytes.
 *     count: Number of vertices.
 * [Return value]
 *     Primitive ID (nonzero), or zero on error.
 */
extern int graphics_create_primitive(
    GraphicsPrimitiveID type, const void *data, const uint32_t *format,
    int size, int count);

/**
 * graphics_create_indexed_primitive:  Create a new graphics primitive
 * object from preexisting vertex and index data buffers.  The result of
 * rendering an indexed primitive is identical to that of rendering a
 * non-indexed primitive with the vertices ordered as listed by the index
 * array, but using an indexed primitive may be faster if there are a lot
 * of repeated vertices.
 *
 * The "vertex_format" parameter is treated identically to the "format"
 * parameter to graphics_create_primitive().  "index_size" specifies the
 * size of each element of the index array in bytes, and must be 1, 2, or
 * 4 (indicating 8-bit, 16-bit, or 32-bit unsigned integers respectively).
 * Some platforms do not support 32-bit indices.
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*).
 *     vertex_data: Vertex data pointer.
 *     vertex_format: Vertex data format description.
 *     vertex_size: Size of a single vertex, in bytes.
 *     vertex_count: Number of vertices in vertex data buffer.
 *     index_data: Index data pointer.
 *     index_size: Size of an index value, in bytes.
 *     index_count: Number of elements in index array (= number of
 *         vertices to render).
 * [Return value]
 *     Primitive ID (nonzero), or zero on error.
 */
extern int graphics_create_indexed_primitive(
    GraphicsPrimitiveID type, const void *vertex_data,
    const uint32_t *vertex_format, int vertex_size, int vertex_count,
    const void *index_data, int index_size, int index_count);

/**
 * graphics_draw_primitive:  Draw a previously-created primitive.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     primitive: ID of primitive to draw.
 */
extern void graphics_draw_primitive(int primitive);

/**
 * graphics_draw_vertices:  Render a graphics primitive from a preexisting
 * vertex data buffer.  Analogous to graphics_create_primitive(), but this
 * function renders the primitive immediately rather than creating a new
 * primitive object for later rendering.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*).
 *     data: Vertex data pointer.
 *     format: Vertex data format description.
 *     size: Size of a single vertex, in bytes.
 *     count: Number of vertices.
 */
extern void graphics_draw_vertices(
    GraphicsPrimitiveID type, const void *data, const uint32_t *format,
    int size, int count);

/**
 * graphics_draw_indexed_vertices:  Render a graphics primitive from
 * preexisting vertex and index data buffers.  Analogous to
 * graphics_create_indexed_primitive(), but this function renders the
 * primitive immediately rather than creating a new primitive object for
 * later rendering.
 *
 * The behavior of this function is undefined if not called between
 * graphics_start_frame() and graphics_finish_frame().
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*).
 *     vertex_data: Vertex data pointer.
 *     vertex_format: Vertex data format description.
 *     vertex_size: Size of a single vertex, in bytes.
 *     vertex_count: Number of vertices in vertex data buffer.
 *     index_data: Index data pointer.
 *     index_size: Size of an index value, in bytes.
 *     index_count: Number of elements in index array (= number of
 *         vertices to render.
 */
extern void graphics_draw_indexed_vertices(
    GraphicsPrimitiveID type, const void *vertex_data,
    const uint32_t *vertex_format, int vertex_size, int vertex_count,
    const void *index_data, int index_size, int index_count);

/**
 * graphics_destroy_primitive:  Destroy a primitive.  Does nothing if
 * primitive == 0.
 *
 * [Parameters]
 *     primitive: ID of primitive to destroy.
 */
extern void graphics_destroy_primitive(int primitive);

/*--------------- Coordinate transformation and clipping ----------------*/

/**
 * graphics_set_viewport:  Set the render viewport position and size.
 * This controls the mapping of transformed (device space) coordinates to
 * window coordinates, but does not affect calls which operate directly on
 * the display or framebuffer (such as graphics_clear() or texture_grab()).
 *
 * All parameters are in units of framebuffer pixels.
 *
 * [Parameters]
 *     left: Leftmost X coordinate of the viewport.
 *     bottom: Bottommost Y coordinate of the viewport.
 *     width: Width of the viewport.
 *     height: Height of the viewport.
 */
extern void graphics_set_viewport(int left, int bottom, int width, int height);

/**
 * graphics_viewport_width, graphics_viewport_height:  Return the width or
 * height of the render viewport.
 *
 * [Return value]
 *     Render viewport width or height, in framebuffer pixels.
 */
extern int graphics_viewport_width(void);
extern int graphics_viewport_height(void);

/**
 * graphics_set_clip_region:  Set the clipping region for rendering and
 * clearing.  If both width and height are zero, the clipping region is
 * cleared (allowing rendering to the entire display); otherwise, both
 * width and height must be nonzero.  The default is to have no clipping
 * region.
 *
 * All parameters are in units of framebuffer pixels and are independent
 * of the viewport setting.  If the clip region and viewport region do not
 * intersect, nothing will be drawn.
 *
 * [Parameters]
 *     left: Leftmost X coordinate of the clipping region.
 *     bottom: Bottommost Y coordinate of the clipping region.
 *     width: Width of the clipping region.
 *     height: Height of the clipping region.
 */
extern void graphics_set_clip_region(int left, int bottom, int width, int height);

/**
 * graphics_set_depth_range:  Set the depth range mapping, which is used
 * to convert device-space depth coordinates to depth buffer values.  The
 * default setting is equivalent to graphics_set_depth_range(0, 1), which
 * maps depth values to the full range of the depth buffer.
 *
 * [Parameters]
 *     near: Window coordinate of the near clipping plane.
 *     far: Window coordinate of the far clipping plane.
 */
extern void graphics_set_depth_range(float near, float far);

/**
 * graphics_set_projection_matrix:  Set the projection transformation matrix
 * to the given data.  The default is the identity matrix.
 *
 * [Parameters]
 *     matrix: Pointer to matrix data.
 */
extern void graphics_set_projection_matrix(const struct Matrix4f *matrix);

/**
 * graphics_set_view_matrix:  Set the view transformation matrix to the
 * given data.  The default is the identity matrix.
 *
 * [Parameters]
 *     matrix: Pointer to matrix data.
 */
extern void graphics_set_view_matrix(const struct Matrix4f *matrix);

/**
 * graphics_set_model_matrix:  Set the model transformation matrix to the
 * given data.  The default is the identity matrix.
 *
 * [Parameters]
 *     matrix: Pointer to matrix data.
 */
extern void graphics_set_model_matrix(const struct Matrix4f *matrix);

/**
 * graphics_get_projection_matrix:  Retrieve the current contents of the
 * projection transformation matrix.
 *
 * [Parameters]
 *     matrix_ret: Pointer to variable to receive matrix data.
 */
extern void graphics_get_projection_matrix(struct Matrix4f *matrix_ret);

/**
 * graphics_get_view_matrix:  Retrieve the current contents of the view
 * transformation matrix.
 *
 * [Parameters]
 *     matrix_ret: Pointer to variable to receive matrix data.
 */
extern void graphics_get_view_matrix(struct Matrix4f *matrix_ret);

/**
 * graphics_get_model_matrix:  Retrieve the current contents of the view
 * transformation matrix.
 *
 * [Parameters]
 *     matrix_ret: Pointer to variable to receive matrix data.
 */
extern void graphics_get_model_matrix(struct Matrix4f *matrix_ret);

/**
 * graphics_set_parallel_projection:  Sets the projection matrix to a
 * parallel transformation matrix defined by the parameters to the
 * function.  Behavior is undefined if any of the three coordinate ranges
 * (right - left, top - bottom, or far - near) is zero.
 *
 * [Parameters]
 *     left: X coordinate for the left edge of the display.
 *     right: X coordinate for the right edge of the display.
 *     bottom: Y coordinate for the bottom edge of the display.
 *     top: Y coordinate for the top edge of the display.
 *     near: Z coordinate for the near clipping plane.
 *     far: Z coordinate for the far clipping plane.
 */
extern void graphics_set_parallel_projection(
    float left, float right, float bottom, float top, float near, float far);

/**
 * graphics_set_perspective_projection:  Sets the projection matrix to a
 * perspective transformation matrix defined by the parameters to the
 * function.  Behavior is undefined if any parameter's value is zero or
 * if far is equal to near.
 *
 * [Parameters]
 *     y_fov: Field-of-view angle in the Y (vertical) direction, in degrees.
 *     aspect: Display aspect ratio (width divided by height).
 *     near: Z coordinate for the near clipping plane.
 *     far: Z coordinate for the far clipping plane.
 */
extern void graphics_set_perspective_projection(
    float y_fov, float aspect, float near, float far);

/*------------------- Other render state manipulation -------------------*/

/**
 * graphics_enable_alpha_test:  Enable or disable alpha testing.  The
 * default is disabled (alpha testing is not performed).
 *
 * [Parameters]
 *     on: True to enable, false to disable.
 */
extern void graphics_enable_alpha_test(int on);

/**
 * graphics_set_alpha_test_comparison:  Set the type of comparison to use
 * for alpha testing.  The default is GRAPHICS_COMPARISON_GREATER_EQUAL,
 * meaning that pixels pass the alpha test if their alpha values are greater
 * than or equal to the reference value.
 *
 * The only permitted comparisons are LESS, LESS_EQUAL, GREATER, and
 * GREATER_EQUAL.
 *
 * [Parameters]
 *     type: Comparison type to use.
 */
extern void graphics_set_alpha_test_comparison(GraphicsComparisonType type);

/**
 * graphics_set_alpha_test_reference:  Set the reference value for alpha
 * testing.  The value is clamped to the inclusive range [0,1].  The
 * default is 0.
 *
 * [Parameters]
 *     alpha: Alpha reference value.
 */
extern void graphics_set_alpha_test_reference(float alpha);

/**
 * graphics_set_blend:  Set the pixel blending operation and pixel factors.
 * The default is to perform standard alpha blending, equivalent to:
 *     graphics_set_blend(GRAPHICS_BLEND_ADD,
 *                        GRAPHICS_BLEND_SRC_ALPHA,
 *                        GRAPHICS_BLEND_INV_SRC_ALPHA)
 *
 * [Parameters]
 *     operation: Blending operation (GRAPHICS_BLEND_ADD, etc.).
 *     src_factor: Source (new) pixel blending factor (GRAPHICS_BLEND_*).
 *     dest_factor: Destination (old) pixel blending factor (GRAPHICS_BLEND_*).
 */
extern void graphics_set_blend(GraphicsBlendOperation operation,
                               GraphicsBlendFactor src_factor,
                               GraphicsBlendFactor dest_factor);

/**
 * graphics_set_blend_color:  Set the color used with the BLEND_CONSTANT
 * and BLEND_INV_CONSTANT blending modes.  Color components are clamped to
 * [0,1].  The default value is (0,0,0,0), equivalent to GRAPHICS_BLEND_ZERO.
 *
 * [Parameters]
 *     color: Blend color.
 */
extern void graphics_set_blend_color(const struct Vector4f *color);

/**
 * graphics_enable_color_write:  Enable or disable writing to the color
 * buffer.  Each color component can be enabled or disabled independently.
 * The default is enabled for all components.
 *
 * [Parameters]
 *     red: True to enable writing to the red component, false to disable.
 *     green: True to enable writing to the green component, false to disable.
 *     blue: True to enable writing to the blue component, false to disable.
 */
extern void graphics_enable_color_write(int red, int green, int blue);

/**
 * graphics_enable_depth_test:  Enable or disable depth testing.  The
 * default is disabled.
 *
 * [Parameters]
 *     on: True to enable, false to disable.
 */
extern void graphics_enable_depth_test(int on);

/**
 * graphics_set_depth_test_comparison:  Set the type of comparison to use
 * for depth testing.  The default is GRAPHICS_COMPARISON_LESS, meaning
 * that pixels pass the depth test if their depth values are strictly less
 * than (closer to the viewer, under the default coordinate transformations)
 * the corresponding value stored in the depth buffer.
 *
 * The only permitted comparisons are LESS, LESS_EQUAL, GREATER, and
 * GREATER_EQUAL.
 *
 * [Parameters]
 *     type: Comparison type to use.
 */
extern void graphics_set_depth_test_comparison(GraphicsComparisonType type);

/**
 * graphics_enable_depth_write:  Enable or disable updating of the depth
 * buffer when rendering.  Regardless of this setting, depth values are not
 * updated if depth testing is disabled.  The default is enabled.
 *
 * [Parameters]
 *     on: True to enable, false to disable.
 */
extern void graphics_enable_depth_write(int on);

/**
 * graphics_set_face_cull:  Set the face culling mode.  The default is
 * GRAPHICS_FACE_CULL_NONE (face culling is not performed).
 *
 * [Parameters]
 *     mode: Face culling mode.
 */
extern void graphics_set_face_cull(GraphicsFaceCullMode mode);

/**
 * graphics_set_fixed_color:  Set the fixed color value to apply to
 * primitives.  Each vertex's color is multiplied by this color when
 * rendering.  Color components are clamped to the range [0,1].  The
 * default value is (1,1,1,1) -- i.e., no color change.
 *
 * [Parameters]
 *     color: Fixed primitive color.
 */
extern void graphics_set_fixed_color(const struct Vector4f *color);

/**
 * graphics_enable_fog:  Enable or disable fog.  Only linear fog is
 * currently supported; the behavior of fog on fragments behind the eye
 * origin is undefined.  The default is disabled.
 *
 * [Parameters]
 *     on: True to enable, false to disable.
 */
extern void graphics_enable_fog(int on);

/**
 * graphics_set_fog_start:  Set the distance at which linear fog is first
 * applied.  The default is 0.
 *
 * [Parameters]
 *     distance: Fog start distance.
 */
extern void graphics_set_fog_start(float distance);

/**
 * graphics_set_fog_end:  Set the distance at which linear fog is fully
 * applied.  The default is 1.
 *
 * Attempting to render a primitive with fog enabled and start == end
 * results in undefined behavior.
 *
 * [Parameters]
 *     distance: Fog end distance.
 */
extern void graphics_set_fog_end(float distance);

/**
 * graphics_set_fog_color:  Set the fog color (the alpha component is
 * ignored).  Color component values are clamped to the range [0,1].
 * The default is (1,1,1,1).
 *
 * [Parameters]
 *     color: Fog color.
 */
extern void graphics_set_fog_color(const struct Vector4f *color);

/**
 * graphics_set_line_width:  Set the line width used for line primitives.
 * The default is 1.
 *
 * [Parameters]
 *     width: Line width, in pixels.
 */
extern void graphics_set_line_width(float width);

/**
 * graphics_set_point_size:  Set the point diameter used for point
 * primitives.  The default is 1.
 *
 * [Parameters]
 *     size: Point diameter, in pixels.
 */
extern void graphics_set_point_size(float size);

/**
 * graphics_enable_stencil_test:  Enable or disable stencil testing.  The
 * default is disabled.
 *
 * [Parameters]
 *     on: True to enable, false to disable.
 */
extern void graphics_enable_stencil_test(int on);

/**
 * graphics_set_stencil_comparison:  Set the stencil comparison type and
 * reference value.  The defaults are:
 *    type = GRAPHICS_COMPARISON_TRUE (stencil test always passes)
 *    reference = 0
 *    mask = ~0
 *
 * Comparisons are performed as:
 *    (stencil_buffer_value & mask) <=> (reference & mask)
 * so that, for example, GRAPHICS_COMPARISON_LESS passes pixels for which
 * the masked stencil buffer value at that pixel is less than the masked
 * reference value.  Note that this is the converse of OpenGL, which puts
 * the reference value on the left side of the comparison.
 *
 * [Parameters]
 *     type: Comparison type to use.
 *     reference: Reference value for comparison.
 *     mask: Bitmask to apply to input stencil buffer value and
 *         reference value.
 */
extern void graphics_set_stencil_comparison(
    GraphicsComparisonType type, unsigned int reference, unsigned int mask);

/**
 * graphics_set_stencil_operations:  Set the stencil operations to be
 * performed based on whether each pixel passes the stencil and depth tests.
 * The default for all operations is GRAPHICS_STENCIL_KEEP.
 *
 * [Parameters]
 *     sfail: Operation to perform when the stencil test fails.
 *     dfail: Operation to perform when the stencil test passes but the
 *         depth test fails.
 *     dpass: Operation to perform when the stencil test passes and the
 *         depth test passes or is disabled.
 */
extern void graphics_set_stencil_operations(
    GraphicsStencilOp sfail, GraphicsStencilOp dfail, GraphicsStencilOp dpass);

/**
 * graphics_set_texture_offset:  Set the offset to add to texture
 * coordinates when rendering textured primitives.  The default is no
 * offset (both components zero).
 *
 * [Parameters]
 *     offset: Texture coordinate offset.
 */
extern void graphics_set_texture_offset(const struct Vector2f *offset);

/*************************************************************************/
/***************************** C++ wrappers ******************************/
/*************************************************************************/

#ifdef __cplusplus
EXTERN_C_END

/*
 * The following macro wraps a function taking an input pointer parameter
 * such that the function can also be called (from C++ only) using a
 * non-pointer value of the same type.  For example:
 *    void graphics_set_projection_matrix(const struct Matrix4f *matrix)
 * is wrapped by:
 *    void graphics_set_projection_matrix(const struct Matrix4f &matrix)
 * allowing a call of the format:
 *    graphics_set_projection_matrix(Matrix4f(...));
 *
 * This is intended mainly to work around the lack of C99-style compound
 * literals in C++.
 */

#define WRAP_FUNCTION(NAME,TYPE) \
    static inline void NAME(const struct TYPE &value) {NAME(&value);}

WRAP_FUNCTION(graphics_set_projection_matrix, Matrix4f)
WRAP_FUNCTION(graphics_set_view_matrix,       Matrix4f)
WRAP_FUNCTION(graphics_set_model_matrix,      Matrix4f)
WRAP_FUNCTION(graphics_set_blend_color,       Vector4f)
WRAP_FUNCTION(graphics_set_fixed_color,       Vector4f)
WRAP_FUNCTION(graphics_set_fog_color,         Vector4f)
WRAP_FUNCTION(graphics_set_texture_offset,    Vector2f)

#undef WRAP_FUNCTION

EXTERN_C_BEGIN
#endif  // __cplusplus

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_GRAPHICS_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
