/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/utility/log.c: Implementation of the DLOG() macro.
 */

#ifdef DEBUG  // To the end of the file.

#include "src/base.h"
#include "src/sysdep.h"
#include "src/utility/log.h"

#ifdef SIL_DLOG_LOG_FILE
# include "src/userdata.h"  // For userdata_get_data_path().
#endif

#ifndef SIL_DLOG_CUSTOM
# include <stdio.h>
#endif

/*************************************************************************/
/****************** Global data (only used for testing) ******************/
/*************************************************************************/

#ifdef INCLUDE_TESTS
char test_DLOG_last_message[SIL_DLOG_MAX_SIZE];
#endif

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void do_DLOG(const char *file, unsigned int line, const char *function,
             const char *format, ...)
{
    va_list args;
    va_start(args, format);
    vdo_DLOG(file, line, function, format, args);
    va_end(args);
}

/*-----------------------------------------------------------------------*/

void vdo_DLOG(const char *file, unsigned int line, const char *function,
             const char *format, va_list args)
{
    char buf[SIL_DLOG_MAX_SIZE];
    int len;

    /* Generate the message header. */
    if (file) {
#ifdef SIL_DLOG_STRIP_PATH
        const unsigned int striplen = strlen(SIL_DLOG_STRIP_PATH);
        if (strncmp(file, SIL_DLOG_STRIP_PATH, striplen) == 0) {
            file += striplen;
        }
#endif
        len = strformat(buf, sizeof(buf), "%s:%u(%s): ", file, line, function);
        len = ubound(len, (int)sizeof(buf)-1);
    } else {
        /* No line header requested, so proceed with an empty buffer. */
        len = 0;
    }

    /* Append the caller-specified message. */
    len += vstrformat(buf + len, sizeof(buf) - len, format, args);
    len = ubound(len, (int)sizeof(buf)-1);
    if (len > 0 && buf[len-1] == '\n') {
        buf[--len] = 0;
    }

#ifdef INCLUDE_TESTS
    memcpy(test_DLOG_last_message, buf, len+1);
#endif

#ifdef SIL_DLOG_CUSTOM

    sys_write_DLOG(buf, len);

#else

    buf[len] = '\n';  // Safe because it was previously a null terminator.

# ifdef SIL_DLOG_LOG_FILE
    {
        static uint8_t opened = 0;
        static FILE *logfp = NULL;

        /* This is technically a race condition, but normally we'll have
         * logged something before any threads start up, so we don't worry
         * about locking. */
        if (!opened && userdata_get_data_path()) {
            opened = 1; // Set this first, to make sure recursive DLOG()
                        // calls from userdata code don't come in here.
            /* Get the directories created if necessary. */
            int id = userdata_save_data("debug.log", "", 0);
            if (id) {
                userdata_wait(id);
                (void) userdata_get_result(id);
            }
            char pathbuf[1000];
            if (strformat_check(pathbuf, sizeof(pathbuf), "%s%s",
                                userdata_get_data_path(),
                                SIL_DLOG_LOG_FILE)) {
                logfp = fopen(pathbuf, "a");
                if (logfp) {
                    setvbuf(logfp, NULL, _IOLBF, 0);  // Line-buffered.
                }
            }
        }

        if (logfp) {
            fwrite(buf, 1, len+1, logfp);
        }
    }
# endif  // SIL_DLOG_LOG_FILE

    fwrite(buf, 1, len+1, stderr);

#endif  // SIL_DLOG_CUSTOM
}

/*************************************************************************/
/*************************************************************************/

#endif  // DEBUG

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
